@@ -10,6 +10,7 @@
 
 #include <spa/pod/filter.h>
 #include <spa/utils/string.h>
+#include <spa/utils/result.h>
 #include <spa/support/system.h>
 #include <spa/utils/keys.h>
 
@@ -442,9 +443,31 @@ int spa_alsa_parse_prop_params(struct state *state, struct spa_pod *params)
 	return changed;
 }
 
+#define CHECK(s,msg,...) if ((err = (s)) < 0) { spa_log_error(state->log, msg ": %s", ##__VA_ARGS__, snd_strerror(err)); return err; }
+
+static ssize_t log_write(void *cookie, const char *buf, size_t size)
+{
+	struct state *state = cookie;
+	int len;
+
+	while (size > 0) {
+		len = strcspn(buf, "\n");
+		if (len > 0)
+			spa_log_debug(state->log, "%.*s", (int)len, buf);
+		buf += len + 1;
+		size -= len + 1;
+	}
+	return size;
+}
+
+static cookie_io_functions_t io_funcs = {
+	.write = log_write,
+};
+
 int spa_alsa_init(struct state *state, const struct spa_dict *info)
 {
 	uint32_t i;
+	int err;
 
 	snd_config_update_free_global();
 
@@ -481,20 +504,34 @@ int spa_alsa_init(struct state *state, const struct spa_dict *info)
 		spa_log_error(state->log, "can't create card %u", state->card_index);
 		return -errno;
 	}
+	state->log_file = fopencookie(state, "w", io_funcs);
+	if (state->log_file == NULL) {
+		spa_log_error(state->log, "can't create log file");
+		return -errno;
+	}
+	CHECK(snd_output_stdio_attach(&state->output, state->log_file, 0), "attach failed");
+
+	state->rate_limit.interval = 2 * SPA_NSEC_PER_SEC;
+	state->rate_limit.burst = 1;
+
 	return 0;
 }
 
 int spa_alsa_clear(struct state *state)
 {
+	int err;
+
 	release_card(state->card);
 
 	state->card = NULL;
 	state->card_index = SPA_ID_INVALID;
 
-	return 0;
-}
+	if ((err = snd_output_close(state->output)) < 0)
+		spa_log_warn(state->log, "output close failed: %s", snd_strerror(err));
+	fclose(state->log_file);
 
-#define CHECK(s,msg,...) if ((err = (s)) < 0) { spa_log_error(state->log, msg ": %s", ##__VA_ARGS__, snd_strerror(err)); return err; }
+	return err;
+}
 
 int spa_alsa_open(struct state *state, const char *params)
 {
@@ -505,8 +542,6 @@ int spa_alsa_open(struct state *state, const char *params)
 	if (state->opened)
 		return 0;
 
-	CHECK(snd_output_stdio_attach(&state->output, stderr, 0), "attach failed");
-
 	spa_scnprintf(device_name, sizeof(device_name), "%s%s%s",
 			state->card->ucm_prefix ? state->card->ucm_prefix : "",
 			props->device, params ? params : "");
@@ -538,6 +573,8 @@ int spa_alsa_open(struct state *state, const char *params)
 	return 0;
 
 error_exit_close:
+	spa_log_info(state->log, "%p: Device '%s' closing: %s", state, state->props.device,
+			spa_strerror(err));
 	snd_pcm_close(state->hndl);
 	return err;
 }
@@ -556,9 +593,6 @@ int spa_alsa_close(struct state *state)
 		spa_log_warn(state->log, "%s: close failed: %s", state->props.device,
 				snd_strerror(err));
 
-	if ((err = snd_output_close(state->output)) < 0)
-		spa_log_warn(state->log, "output close failed: %s", snd_strerror(err));
-
 	spa_system_close(state->data_system, state->timerfd);
 
 	if (state->have_format)
@@ -604,12 +638,10 @@ static const struct format_info format_info[] = {
 
 static snd_pcm_format_t spa_format_to_alsa(uint32_t format, bool *planar)
 {
-	size_t i;
-
-	for (i = 0; i < SPA_N_ELEMENTS(format_info); i++) {
-		*planar = format_info[i].spa_pformat == format;
-		if (format_info[i].spa_format == format || *planar)
-			return format_info[i].format;
+	SPA_FOR_EACH_ELEMENT_VAR(format_info, i) {
+		*planar = i->spa_pformat == format;
+		if (i->spa_format == format || *planar)
+			return i->format;
 	}
 	return SND_PCM_FORMAT_UNKNOWN;
 }
@@ -751,8 +783,8 @@ static bool uint32_array_contains(uint32_t *vals, uint32_t n_vals, uint32_t val)
 	return false;
 }
 
-static int add_rate(struct state *state, uint32_t scale, bool all, uint32_t index, uint32_t *next,
-		snd_pcm_hw_params_t *params, struct spa_pod_builder *b)
+static int add_rate(struct state *state, uint32_t scale, uint32_t interleave, bool all, uint32_t index, uint32_t *next,
+		uint32_t min_allowed_rate, snd_pcm_hw_params_t *params, struct spa_pod_builder *b)
 {
 	struct spa_pod_frame f[1];
 	int err, dir;
@@ -763,6 +795,14 @@ static int add_rate(struct state *state, uint32_t scale, bool all, uint32_t inde
 	CHECK(snd_pcm_hw_params_get_rate_min(params, &min, &dir), "get_rate_min");
 	CHECK(snd_pcm_hw_params_get_rate_max(params, &max, &dir), "get_rate_max");
 
+	spa_log_debug(state->log, "min:%u max:%u min-allowed:%u scale:%u interleave:%u all:%d",
+			min, max, min_allowed_rate, scale, interleave, all);
+
+	min = SPA_MAX(min_allowed_rate * scale / interleave, min) * interleave / scale;
+	max = max * interleave / scale;
+	if (max < min)
+		return 0;
+
 	if (!state->multi_rate && state->card->format_ref > 0)
 		rate = state->card->rate;
 	else
@@ -779,6 +819,9 @@ static int add_rate(struct state *state, uint32_t scale, bool all, uint32_t inde
 
 	rate = SPA_CLAMP(rate, min, max);
 
+	spa_log_debug(state->log, "rate:%u multi:%d card:%d def:%d",
+			rate, state->multi_rate, state->card->rate, state->default_rate);
+
 	spa_pod_builder_prop(b, SPA_FORMAT_AUDIO_rate, 0);
 
 	spa_pod_builder_push_choice(b, &f[0], SPA_CHOICE_None, 0);
@@ -830,7 +873,8 @@ static int add_channels(struct state *state, bool all, uint32_t index, uint32_t
 
 	CHECK(snd_pcm_hw_params_get_channels_min(params, &min), "get_channels_min");
 	CHECK(snd_pcm_hw_params_get_channels_max(params, &max), "get_channels_max");
-	spa_log_debug(state->log, "channels (%d %d)", min, max);
+	spa_log_debug(state->log, "channels (%d %d) default:%d all:%d",
+			min, max, state->default_channels, all);
 
 	if (state->default_channels != 0 && !all) {
 		if (min < state->default_channels)
@@ -911,11 +955,19 @@ skip_channels:
 	return 1;
 }
 
+static void debug_hw_params(struct state *state, const char *prefix, snd_pcm_hw_params_t *params)
+{
+	if (SPA_UNLIKELY(spa_log_level_topic_enabled(state->log, SPA_LOG_TOPIC_DEFAULT, SPA_LOG_LEVEL_DEBUG))) {
+		spa_log_debug(state->log, "%s:", prefix);
+		snd_pcm_hw_params_dump(params, state->output);
+		fflush(state->log_file);
+	}
+}
 static int enum_pcm_formats(struct state *state, uint32_t index, uint32_t *next,
 		struct spa_pod **result, struct spa_pod_builder *b)
 {
 	int res, err;
-	size_t i, j;
+	size_t j;
 	snd_pcm_t *hndl;
 	snd_pcm_hw_params_t *params;
 	struct spa_pod_frame f[2];
@@ -928,6 +980,8 @@ static int enum_pcm_formats(struct state *state, uint32_t index, uint32_t *next,
 	snd_pcm_hw_params_alloca(&params);
 	CHECK(snd_pcm_hw_params_any(hndl, params), "Broken configuration: no configurations available");
 
+	debug_hw_params(state, __func__, params);
+
 	CHECK(snd_pcm_hw_params_set_rate_resample(hndl, params, 0), "set_rate_resample");
 
 	if (state->default_channels != 0) {
@@ -964,8 +1018,10 @@ static int enum_pcm_formats(struct state *state, uint32_t index, uint32_t *next,
 	spa_pod_builder_push_choice(b, &f[1], SPA_CHOICE_None, 0);
 	choice = (struct spa_pod_choice*)spa_pod_builder_frame(b, &f[1]);
 
-	for (i = 1, j = 0; i < SPA_N_ELEMENTS(format_info); i++) {
-		const struct format_info *fi = &format_info[i];
+	j = 0;
+	SPA_FOR_EACH_ELEMENT_VAR(format_info, fi) {
+		if (fi->format == SND_PCM_FORMAT_UNKNOWN)
+			continue;
 
 		if (snd_pcm_format_mask_test(fmask, fi->format)) {
 			if ((snd_pcm_access_mask_test(amask, SND_PCM_ACCESS_MMAP_NONINTERLEAVED) ||
@@ -1017,7 +1073,7 @@ static int enum_pcm_formats(struct state *state, uint32_t index, uint32_t *next,
 		choice->body.type = SPA_CHOICE_Enum;
 	spa_pod_builder_pop(b, &f[1]);
 
-	if ((res = add_rate(state, 1, false, index & 0xffff, next, params, b)) != 1)
+	if ((res = add_rate(state, 1, 1, false, index & 0xffff, next, 0, params, b)) != 1)
 		return res;
 
 	if ((res = add_channels(state, false, index & 0xffff, next, params, b)) != 1)
@@ -1074,6 +1130,8 @@ static int enum_iec958_formats(struct state *state, uint32_t index, uint32_t *ne
 	snd_pcm_hw_params_alloca(&params);
 	CHECK(snd_pcm_hw_params_any(hndl, params), "Broken configuration: no configurations available");
 
+	debug_hw_params(state, __func__, params);
+
 	CHECK(snd_pcm_hw_params_set_rate_resample(hndl, params, 0), "set_rate_resample");
 
 	spa_pod_builder_push_object(b, &f[0], SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat);
@@ -1110,7 +1168,7 @@ static int enum_iec958_formats(struct state *state, uint32_t index, uint32_t *ne
 	}
 	spa_pod_builder_pop(b, &f[1]);
 
-	if ((res = add_rate(state, 1, true, index & 0xffff, next, params, b)) != 1)
+	if ((res = add_rate(state, 1, 1, true, index & 0xffff, next, 0, params, b)) != 1)
 		return res;
 
 	(*next)++;
@@ -1135,6 +1193,8 @@ static int enum_dsd_formats(struct state *state, uint32_t index, uint32_t *next,
 	snd_pcm_hw_params_alloca(&params);
 	CHECK(snd_pcm_hw_params_any(hndl, params), "Broken configuration: no configurations available");
 
+	debug_hw_params(state, __func__, params);
+
 	snd_pcm_format_mask_alloca(&fmask);
 	snd_pcm_hw_params_get_format_mask(params, fmask);
 
@@ -1165,7 +1225,15 @@ static int enum_dsd_formats(struct state *state, uint32_t index, uint32_t *next,
 	spa_pod_builder_prop(b, SPA_FORMAT_AUDIO_interleave, 0);
 	spa_pod_builder_int(b, interleave);
 
-	if ((res = add_rate(state, SPA_ABS(interleave), true, index & 0xffff, next, params, b)) != 1)
+	/* Use a lower rate limit of 352800 (= 44100 * 64 / 8). This is because in
+	 * PipeWire, DSD rates are given in bytes, not bits, so 352800 corresponds
+	 * to the bit rate of DSD64. (The "64" in DSD64 means "64 times the rate
+	 * of 44.1 kHz".) Some hardware may report rates lower than that, for example
+	 * 176400. This would correspond to "DSD32" (which does not exist). Trying
+	 * to use such a rate with DSD hardware does not work and may cause undefined
+	 * behavior in said hardware. */
+	if ((res = add_rate(state, 8, SPA_ABS(interleave), true, index & 0xffff,
+					next, 44100, params, b)) != 1)
 		return res;
 
 	if ((res = add_channels(state, true, index & 0xffff, next, params, b)) != 1)
@@ -1187,7 +1255,12 @@ spa_alsa_enum_format(struct state *state, int seq, uint32_t start, uint32_t num,
 	struct spa_result_node_params result;
 	uint32_t count = 0;
 
+	spa_log_debug(state->log, "opened:%d format:%d started:%d", state->opened,
+			state->have_format, state->started);
+
 	opened = state->opened;
+	if (!state->started && state->have_format)
+		spa_alsa_close(state);
 	if ((err = spa_alsa_open(state, NULL)) < 0)
 		return err;
 
@@ -1248,6 +1321,9 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 	bool match = true, planar = false, is_batch;
 	char spdif_params[128] = "";
 
+	spa_log_debug(state->log, "opened:%d format:%d started:%d", state->opened,
+			state->have_format, state->started);
+
 	state->use_mmap = !state->disable_mmap;
 
 	switch (fmt->media_subtype) {
@@ -1360,6 +1436,8 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 		return -EINVAL;
 	}
 
+	if (!state->started && state->have_format)
+		spa_alsa_close(state);
 	if ((err = spa_alsa_open(state, spdif_params)) < 0)
 		return err;
 
@@ -1368,6 +1446,9 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 	snd_pcm_hw_params_alloca(&params);
 	/* choose all parameters */
 	CHECK(snd_pcm_hw_params_any(hndl, params), "Broken configuration for playback: no configurations available");
+
+	debug_hw_params(state, __func__, params);
+
 	/* set hardware resampling, no resample */
 	CHECK(snd_pcm_hw_params_set_rate_resample(hndl, params, 0), "set_rate_resample");
 
@@ -1555,6 +1636,12 @@ static int set_swparams(struct state *state)
 	/* write the parameters to the playback device */
 	CHECK(snd_pcm_sw_params(hndl, params), "sw_params");
 
+	if (SPA_UNLIKELY(spa_log_level_topic_enabled(state->log, SPA_LOG_TOPIC_DEFAULT, SPA_LOG_LEVEL_DEBUG))) {
+		spa_log_debug(state->log, "state after sw_params:");
+		snd_pcm_dump(hndl, state->output);
+		fflush(state->log_file);
+	}
+
 	return 0;
 }
 
@@ -1688,22 +1775,24 @@ recover:
 	state->alsa_started = false;
 
 	if (state->stream == SND_PCM_STREAM_PLAYBACK)
-		spa_alsa_silence(state, state->start_delay + state->threshold * 2 + state->headroom);
+		spa_alsa_silence(state, state->start_delay + state->threshold + state->headroom);
 
 	return do_start(state);
 }
 
 static int get_avail(struct state *state, uint64_t current_time)
 {
-	int res;
+	int res, missed;
 	snd_pcm_sframes_t avail;
 
 	if (SPA_UNLIKELY((avail = snd_pcm_avail(state->hndl)) < 0)) {
 		if ((res = alsa_recover(state, avail)) < 0)
 			return res;
 		if ((avail = snd_pcm_avail(state->hndl)) < 0) {
-			spa_log_warn(state->log, "%s: snd_pcm_avail after recover: %s",
-					state->props.device, snd_strerror(avail));
+			if ((missed = ratelimit_test(&state->rate_limit, current_time)) >= 0) {
+				spa_log_warn(state->log, "%s: (%d missed) snd_pcm_avail after recover: %s",
+						state->props.device, missed, snd_strerror(avail));
+			}
 			avail = state->threshold * 2;
 		}
 	} else {
@@ -1715,7 +1804,7 @@ static int get_avail(struct state *state, uint64_t current_time)
 #if 0
 static int get_avail_htimestamp(struct state *state, uint64_t current_time)
 {
-	int res;
+	int res, missed;
 	snd_pcm_uframes_t avail;
 	snd_htimestamp_t tstamp;
 	uint64_t then;
@@ -1724,8 +1813,10 @@ static int get_avail_htimestamp(struct state *state, uint64_t current_time)
 		if ((res = alsa_recover(state, avail)) < 0)
 			return res;
 		if ((res = snd_pcm_htimestamp(state->hndl, &avail, &tstamp)) < 0) {
-			spa_log_warn(state->log, "%s: snd_pcm_htimestamp error: %s",
-				state->props.device, snd_strerror(res));
+			if ((missed = ratelimit_test(&state->rate_limit, current_time)) >= 0) {
+				spa_log_warn(state->log, "%s: (%d missed) snd_pcm_htimestamp error: %s",
+					state->props.device, missed, snd_strerror(res));
+			}
 			avail = state->threshold * 2;
 		}
 	} else {
@@ -1768,6 +1859,7 @@ static int get_status(struct state *state, uint64_t current_time,
 		*delay = avail;
 		*target = SPA_MAX(*target, state->read_size);
 	}
+	*target = SPA_MIN(*target, state->buffer_frames);
 	return 0;
 }
 
@@ -1795,15 +1887,14 @@ static int update_time(struct state *state, uint64_t current_time, snd_pcm_sfram
 				state, follower, state->last_threshold, state->threshold, diff, err);
 		state->last_threshold = state->threshold;
 		state->alsa_sync = true;
+		state->alsa_sync_warning = false;
 	}
 	if (err > state->max_error) {
 		err = state->max_error;
 		state->alsa_sync = true;
-		state->alsa_sync_warning = (diff == 0);
 	} else if (err < -state->max_error) {
 		err = -state->max_error;
 		state->alsa_sync = true;
-		state->alsa_sync_warning = (diff == 0);
 	}
 
 	if (!follower || state->matching)
@@ -1900,7 +1991,7 @@ int spa_alsa_write(struct state *state)
 	const snd_pcm_channel_area_t *my_areas;
 	snd_pcm_uframes_t written, frames, offset, off, to_write, total_written, max_write;
 	snd_pcm_sframes_t commitres;
-	int res = 0;
+	int res = 0, missed;
 	size_t frame_size = state->frame_size;
 
 	check_position_config(state);
@@ -1920,13 +2011,18 @@ int spa_alsa_write(struct state *state)
 			return res;
 
 		if (SPA_UNLIKELY(state->alsa_sync)) {
-			if (SPA_UNLIKELY(state->alsa_sync_warning)) {
-				spa_log_warn(state->log, "%s: follower delay:%ld target:%ld thr:%u, resync",
-						state->props.device, delay, target, state->threshold);
-				state->alsa_sync_warning = false;
-			} else
-				spa_log_info(state->log, "%s: follower delay:%ld target:%ld thr:%u, resync",
-						state->props.device, delay, target, state->threshold);
+			enum spa_log_level lev;
+
+			if (SPA_UNLIKELY(state->alsa_sync_warning))
+				lev = SPA_LOG_LEVEL_WARN;
+			else
+				lev = SPA_LOG_LEVEL_INFO;
+
+			if ((missed = ratelimit_test(&state->rate_limit, current_time)) >= 0) {
+				spa_log_lev(state->log, lev, "%s: follower delay:%ld target:%ld thr:%u, "
+						"resync (%d missed)", state->props.device, delay,
+						target, state->threshold, missed);
+			}
 
 			if (delay > target)
 				snd_pcm_rewind(state->hndl, delay - target);
@@ -1934,7 +2030,8 @@ int spa_alsa_write(struct state *state)
 				spa_alsa_silence(state, target - delay);
 			delay = target;
 			state->alsa_sync = false;
-		}
+		} else
+			state->alsa_sync_warning = true;
 	}
 
 	total_written = 0;
@@ -1961,13 +2058,14 @@ again:
 		size_t n_bytes, n_frames;
 		struct buffer *b;
 		struct spa_data *d;
-		uint32_t i, offs, size;
+		uint32_t i, offs, size, last_offset;
 
 		b = spa_list_first(&state->ready, struct buffer, link);
 		d = b->buf->datas;
 
 		offs = d[0].chunk->offset + state->ready_offset;
-		size = d[0].chunk->size - state->ready_offset;
+		last_offset = d[0].chunk->size;
+		size = last_offset - state->ready_offset;
 
 		offs = SPA_MIN(offs, d[0].maxsize);
 		size = SPA_MIN(d[0].maxsize - offs, size);
@@ -1993,7 +2091,7 @@ again:
 
 		state->ready_offset += n_bytes;
 
-		if (state->ready_offset >= size) {
+		if (state->ready_offset >= last_offset) {
 			spa_list_remove(&b->link);
 			SPA_FLAG_SET(b->flags, BUFFER_FLAG_OUT);
 			state->io->buffer_id = b->id;
@@ -2030,7 +2128,7 @@ again:
 
 	state->sample_count += total_written;
 
-	if (SPA_UNLIKELY(!state->alsa_started && total_written > 0))
+	if (SPA_UNLIKELY(!state->alsa_started && (total_written > 0 || frames == 0)))
 		do_start(state);
 
 	return 0;
@@ -2126,7 +2224,7 @@ int spa_alsa_read(struct state *state)
 	const snd_pcm_channel_area_t *my_areas;
 	snd_pcm_uframes_t read, frames, offset;
 	snd_pcm_sframes_t commitres;
-	int res = 0;
+	int res = 0, missed;
 
 	check_position_config(state);
 
@@ -2135,7 +2233,6 @@ int spa_alsa_read(struct state *state)
 	if (state->following && state->alsa_started) {
 		uint64_t current_time;
 		snd_pcm_uframes_t avail, delay, target;
-		uint32_t threshold = state->threshold;
 
 		current_time = state->position->clock.nsec;
 
@@ -2148,13 +2245,18 @@ int spa_alsa_read(struct state *state)
 			return res;
 
 		if (state->alsa_sync) {
-			if (SPA_UNLIKELY(state->alsa_sync_warning)) {
-				spa_log_warn(state->log, "%s: follower delay:%lu target:%lu thr:%u, resync",
-						state->props.device, delay, target, threshold);
-				state->alsa_sync_warning = false;
-			} else
-				spa_log_info(state->log, "%s: follower delay:%lu target:%lu thr:%u, resync",
-						state->props.device, delay, target, threshold);
+			enum spa_log_level lev;
+
+			if (SPA_UNLIKELY(state->alsa_sync_warning))
+				lev = SPA_LOG_LEVEL_WARN;
+			else
+				lev = SPA_LOG_LEVEL_INFO;
+
+			if ((missed = ratelimit_test(&state->rate_limit, current_time)) >= 0) {
+				spa_log_lev(state->log, lev, "%s: follower delay:%ld target:%ld thr:%u, "
+						"resync (%d missed)", state->props.device, delay,
+						target, state->threshold, missed);
+			}
 
 			if (delay < target)
 				max_read = target - delay;
@@ -2162,7 +2264,8 @@ int spa_alsa_read(struct state *state)
 				snd_pcm_forward(state->hndl, delay - target);
 			delay = target;
 			state->alsa_sync = false;
-		}
+		} else
+			state->alsa_sync_warning = true;
 
 		if (avail < state->read_size)
 			max_read = 0;
@@ -2252,7 +2355,7 @@ static int handle_play(struct state *state, uint64_t current_time,
 {
 	int res;
 
-	if (SPA_UNLIKELY(delay > target + state->max_error)) {
+	if (state->alsa_started && SPA_UNLIKELY(delay > target + state->max_error)) {
 		spa_log_trace(state->log, "%p: early wakeup %lu %lu", state, delay, target);
 		if (delay > target * 3)
 			delay = target * 3;
@@ -2340,7 +2443,7 @@ static void alsa_on_timeout_event(struct spa_source *source)
 	}
 
 #ifndef FASTPATH
-	if (SPA_UNLIKELY(spa_log_level_enabled(state->log, SPA_LOG_LEVEL_TRACE))) {
+	if (SPA_UNLIKELY(spa_log_level_topic_enabled(state->log, SPA_LOG_TOPIC_DEFAULT, SPA_LOG_LEVEL_TRACE))) {
 		struct timespec now;
 		uint64_t nsec;
 		if (spa_system_clock_gettime(state->data_system, CLOCK_MONOTONIC, &now) < 0)
@@ -2435,8 +2538,6 @@ int spa_alsa_start(struct state *state)
 			state->following, state->matching, state->resample);
 
 	CHECK(set_swparams(state), "swparams");
-	if (SPA_UNLIKELY(spa_log_level_enabled(state->log, SPA_LOG_LEVEL_DEBUG)))
-		snd_pcm_dump(state->hndl, state->output);
 
 	if ((err = snd_pcm_prepare(state->hndl)) < 0 && err != -EBUSY) {
 		spa_log_error(state->log, "%s: snd_pcm_prepare error: %s",
@@ -2453,13 +2554,14 @@ int spa_alsa_start(struct state *state)
 
 	reset_buffers(state);
 	state->alsa_sync = true;
+	state->alsa_sync_warning = false;
 	state->alsa_recovering = false;
 	state->alsa_started = false;
 
+	/* start capture now, playback will start after first write */
 	if (state->stream == SND_PCM_STREAM_PLAYBACK)
-		spa_alsa_silence(state, state->start_delay + state->threshold * 2 + state->headroom);
-
-	if ((err = do_start(state)) < 0)
+		spa_alsa_silence(state, state->start_delay + state->threshold + state->headroom);
+	else if ((err = do_start(state)) < 0)
 		return err;
 
 	set_timers(state);
