@@ -519,13 +519,13 @@ struct link {
 
 struct graph_port {
 	const struct fc_descriptor *desc;
-	void *hndl;
+	void **hndl;
 	uint32_t port;
 };
 
 struct graph_hndl {
 	const struct fc_descriptor *desc;
-	void *hndl;
+	void **hndl;
 };
 
 struct graph {
@@ -577,6 +577,10 @@ struct impl {
 	struct graph graph;
 };
 
+static int graph_instantiate(struct graph *graph);
+static void graph_cleanup(struct graph *graph);
+
+
 static void capture_destroy(void *d)
 {
 	struct impl *impl = d;
@@ -585,51 +589,72 @@ static void capture_destroy(void *d)
 }
 
 static void capture_process(void *d)
+{
+	struct impl *impl = d;
+	pw_stream_trigger_process(impl->playback);
+}
+
+static void playback_process(void *d)
 {
 	struct impl *impl = d;
 	struct pw_buffer *in, *out;
 	struct graph *graph = &impl->graph;
-	uint32_t i, outsize = 0, n_hndl = graph->n_hndl;
+	uint32_t i, insize = 0, outsize = 0, n_hndl = graph->n_hndl;
 	int32_t stride = 0;
+	struct graph_port *port;
+	struct spa_data *bd;
 
 	if ((in = pw_stream_dequeue_buffer(impl->capture)) == NULL)
-		pw_log_debug("out of capture buffers: %m");
+		pw_log_debug("%p: out of capture buffers: %m", impl);
 
 	if ((out = pw_stream_dequeue_buffer(impl->playback)) == NULL)
-		pw_log_debug("out of playback buffers: %m");
+		pw_log_debug("%p: out of playback buffers: %m", impl);
 
 	if (in == NULL || out == NULL)
 		goto done;
 
 	for (i = 0; i < in->buffer->n_datas; i++) {
-		struct spa_data *ds = &in->buffer->datas[i];
-		struct graph_port *port = &graph->input[i];
 		uint32_t offs, size;
 
-		offs = SPA_MIN(ds->chunk->offset, ds->maxsize);
-		size = SPA_MIN(ds->chunk->size, ds->maxsize - offs);
+		bd = &in->buffer->datas[i];
 
-		if (port->desc)
-			port->desc->connect_port(port->hndl, port->port,
-				SPA_PTROFF(ds->data, offs, void));
+		offs = SPA_MIN(bd->chunk->offset, bd->maxsize);
+		size = SPA_MIN(bd->chunk->size, bd->maxsize - offs);
 
-		outsize = SPA_MAX(outsize, size);
-		stride = SPA_MAX(stride, ds->chunk->stride);
+		port = i < graph->n_input ? &graph->input[i] : NULL;
+
+		if (port && port->desc)
+			port->desc->connect_port(*port->hndl, port->port,
+				SPA_PTROFF(bd->data, offs, void));
+
+		insize = i == 0 ? size : SPA_MIN(insize, size);
+		stride = SPA_MAX(stride, bd->chunk->stride);
 	}
+	outsize = insize;
+
 	for (i = 0; i < out->buffer->n_datas; i++) {
-		struct spa_data *dd = &out->buffer->datas[i];
-		struct graph_port *port = &graph->output[i];
-		if (port->desc)
-			port->desc->connect_port(port->hndl, port->port, dd->data);
+		bd = &out->buffer->datas[i];
+
+		outsize = SPA_MIN(outsize, bd->maxsize);
+
+		port = i < graph->n_output ? &graph->output[i] : NULL;
+
+		if (port && port->desc)
+			port->desc->connect_port(*port->hndl, port->port, bd->data);
 		else
-			memset(dd->data, 0, outsize);
-		dd->chunk->offset = 0;
-		dd->chunk->size = outsize;
-		dd->chunk->stride = stride;
+			memset(bd->data, 0, outsize);
+
+		bd->chunk->offset = 0;
+		bd->chunk->size = outsize;
+		bd->chunk->stride = stride;
 	}
+
+	pw_log_trace_fp("%p: stride:%d in:%d out:%d requested:%"PRIu64" (%"PRIu64")", impl,
+			stride, insize, outsize, out->requested, out->requested * stride);
+
 	for (i = 0; i < n_hndl; i++) {
 		struct graph_hndl *hndl = &graph->hndl[i];
-		hndl->desc->run(hndl->hndl, outsize / sizeof(float));
+		hndl->desc->run(*hndl->hndl, outsize / sizeof(float));
 	}
 
 done:
@@ -637,8 +662,6 @@ done:
 		pw_stream_queue_buffer(impl->capture, in);
 	if (out != NULL)
 		pw_stream_queue_buffer(impl->playback, out);
-
-	pw_stream_trigger_process(impl->playback);
 }
 
 static float get_default(struct impl *impl, struct descriptor *desc, uint32_t p)
@@ -673,11 +696,21 @@ static struct port *find_port(struct node *node, const char *name, int descripto
 	str = strdupa(name);
 	col = strchr(str, ':');
 	if (col != NULL) {
+		struct node *find;
 		node_name = str;
 		port_name = col + 1;
 		*col = '\0';
-		node = find_node(node->graph, node_name);
-	} else {
+		find = find_node(node->graph, node_name);
+		if (find == NULL) {
+			/* it's possible that the : is part of the port name,
+			 * try again without splitting things up. */
+			*col = ':';
+			col = NULL;
+		} else {
+			node = find;
+		}
+	}
+	if (col == NULL) {
 		node_name = node->name;
 		port_name = str;
 	}
@@ -727,11 +760,12 @@ static struct spa_pod *get_prop_info(struct graph *graph, struct spa_pod_builder
 	struct fc_port *p = &d->ports[port->p];
 	float def, min, max;
 	char name[512];
+	uint32_t rate = impl->rate ? impl->rate : 48000;
 
 	if (p->hint & FC_HINT_SAMPLE_RATE) {
-		def = p->def * impl->rate;
-		min = p->min * impl->rate;
-		max = p->max * impl->rate;
+		def = p->def * rate;
+		min = p->min * rate;
+		max = p->max * rate;
 	} else {
 		def = p->def;
 		min = p->min;
@@ -751,10 +785,10 @@ static struct spa_pod *get_prop_info(struct graph *graph, struct spa_pod_builder
 	spa_pod_builder_prop(b, SPA_PROP_INFO_type, 0);
 	if (p->hint & FC_HINT_BOOLEAN) {
 		if (min == max) {
-			spa_pod_builder_bool(b, def <= 0.0 ? false : true);
+			spa_pod_builder_bool(b, def <= 0.0f ? false : true);
 		} else  {
 			spa_pod_builder_push_choice(b, &f[1], SPA_CHOICE_Enum, 0);
-			spa_pod_builder_bool(b, def <= 0.0 ? false : true);
+			spa_pod_builder_bool(b, def <= 0.0f ? false : true);
 			spa_pod_builder_bool(b, false);
 			spa_pod_builder_bool(b, true);
 			spa_pod_builder_pop(b, &f[1]);
@@ -810,7 +844,7 @@ static struct spa_pod *get_props_param(struct graph *graph, struct spa_pod_build
 
 		spa_pod_builder_string(b, name);
 		if (p->hint & FC_HINT_BOOLEAN) {
-			spa_pod_builder_bool(b, port->control_data <= 0.0 ? false : true);
+			spa_pod_builder_bool(b, port->control_data <= 0.0f ? false : true);
 		} else if (p->hint & FC_HINT_INTEGER) {
 			spa_pod_builder_int(b, port->control_data);
 		} else {
@@ -829,7 +863,7 @@ static int set_control_value(struct node *node, const char *name, float *value)
 
 	port = find_port(node, name, FC_PORT_INPUT | FC_PORT_CONTROL);
 	if (port == NULL)
-		return 0;
+		return -ENOENT;
 
 	node = port->node;
 	desc = node->desc;
@@ -844,7 +878,7 @@ static int parse_params(struct graph *graph, const struct spa_pod *pod)
 {
 	struct spa_pod_parser prs;
 	struct spa_pod_frame f;
-	int changed = 0;
+	int res, changed = 0;
 	struct node *def_node;
 
 	def_node = spa_list_first(&graph->node_list, struct node, link);
@@ -877,7 +911,8 @@ static int parse_params(struct graph *graph, const struct spa_pod *pod)
 			struct spa_pod *pod;
 			spa_pod_parser_get_pod(&prs, &pod);
 		}
-		changed += set_control_value(def_node, name, val);
+		if ((res = set_control_value(def_node, name, val)) > 0)
+			changed += res;
 	}
 	return changed;
 }
@@ -888,13 +923,14 @@ static void graph_reset(struct graph *graph)
 	for (i = 0; i < graph->n_hndl; i++) {
 		struct graph_hndl *hndl = &graph->hndl[i];
 		const struct fc_descriptor *d = hndl->desc;
+		if (hndl->hndl == NULL || *hndl->hndl == NULL)
+			continue;
 		if (d->deactivate)
-			d->deactivate(hndl->hndl);
+			d->deactivate(*hndl->hndl);
 		if (d->activate)
-			d->activate(hndl->hndl);
+			d->activate(*hndl->hndl);
 	}
 }
-
 static void param_props_changed(struct impl *impl, const struct spa_pod *param)
 {
 	struct spa_pod_object *obj = (struct spa_pod_object *) param;
@@ -966,11 +1002,25 @@ static void param_changed(void *data, uint32_t id, const struct spa_pod *param)
 {
 	struct impl *impl = data;
 	struct graph *graph = &impl->graph;
+	int res;
 
 	switch (id) {
 	case SPA_PARAM_Format:
-		if (param == NULL)
-			graph_reset(graph);
+		if (param == NULL) {
+			graph_cleanup(graph);
+		} else {
+			struct spa_audio_info_raw info;
+			spa_zero(info);
+			if ((res = spa_format_audio_raw_parse(param, &info)) < 0)
+				goto error;
+			if (info.rate == 0) {
+				res = -EINVAL;
+				goto error;
+			}
+			impl->rate = info.rate;
+			if ((res = graph_instantiate(graph)) < 0)
+				goto error;
+		}
 		break;
 	case SPA_PARAM_Props:
 		if (param != NULL)
@@ -980,6 +1030,11 @@ static void param_changed(void *data, uint32_t id, const struct spa_pod *param)
 		param_latency_changed(impl, param);
 		break;
 	}
+	return;
+
+error:
+	pw_stream_set_error(impl->capture, res, "can't start graph: %s",
+			spa_strerror(res));
 }
 
 static const struct pw_stream_events in_stream_events = {
@@ -1000,6 +1055,7 @@ static void playback_destroy(void *d)
 static const struct pw_stream_events out_stream_events = {
 	PW_VERSION_STREAM_EVENTS,
 	.destroy = playback_destroy,
+	.process = playback_process,
 	.state_changed = state_changed,
 	.param_changed = param_changed
 };
@@ -1128,16 +1184,19 @@ static struct plugin *plugin_load(struct impl *impl, const char *type, const cha
 	else if (spa_streq(type, "ladspa")) {
 		pl = load_ladspa_plugin(support, n_support, path, NULL);
 	}
-#ifdef HAVE_LILV
 	else if (spa_streq(type, "lv2")) {
+#ifdef HAVE_LILV
 		pl = load_lv2_plugin(support, n_support, path, NULL);
-	}
+#else
+		pw_log_error("filter-chain is compiled without lv2 support");
+		pl = NULL;
+		errno = ENOTSUP;
 #endif
-	else {
+	} else {
+		pw_log_error("invalid plugin type '%s'", type);
 		pl = NULL;
 		errno = EINVAL;
 	}
-
 	if (pl == NULL)
 		goto exit;
 
@@ -1329,15 +1388,17 @@ static int parse_control(struct node *node, struct spa_json *control)
 	while (spa_json_get_string(control, key, sizeof(key)) > 0) {
 		float fl;
 		const char *val;
-		int len;
+		int res, len;
 
 		if ((len = spa_json_next(control, &val)) < 0)
 			break;
 
-		if (spa_json_parse_float(val, len, &fl) <= 0)
+		if (spa_json_parse_float(val, len, &fl) <= 0) {
 			pw_log_warn("control '%s' expects a number, ignoring", key);
-		else
-			set_control_value(node, key, &fl);
+		}
+		else if ((res = set_control_value(node, key, &fl)) < 0) {
+			pw_log_warn("control '%s' can not be set: %s", key, spa_strerror(res));
+		}
 	}
 	return 0;
 }
@@ -1492,10 +1553,8 @@ static int load_node(struct graph *graph, struct spa_json *json)
 			break;
 	}
 
-	if (spa_streq(type, "builtin")) {
+	if (spa_streq(type, "builtin"))
 		snprintf(plugin, sizeof(plugin), "%s", "builtin");
-	} else if (!spa_streq(type, "ladspa") && !spa_streq(type, "lv2"))
-		return -ENOTSUP;
 
 	pw_log_info("loading type:%s plugin:%s label:%s", type, plugin, label);
 
@@ -1515,6 +1574,10 @@ static int load_node(struct graph *graph, struct spa_json *json)
 	node->control_port = calloc(desc->n_control, sizeof(struct port));
 	node->notify_port = calloc(desc->n_notify, sizeof(struct port));
 
+	pw_log_info("loaded n_input:%d n_output:%d n_control:%d n_notify:%d",
+			desc->n_input, desc->n_output,
+			desc->n_control, desc->n_notify);
+
 	for (i = 0; i < desc->n_input; i++) {
 		struct port *port = &node->input_port[i];
 		port->node = node;
@@ -1559,21 +1622,51 @@ static int load_node(struct graph *graph, struct spa_json *json)
 	return 0;
 }
 
-static void node_free(struct node *node)
+static void node_cleanup(struct node *node)
 {
-	uint32_t i, j;
 	const struct fc_descriptor *d = node->desc->desc;
+	uint32_t i;
 
-	spa_list_remove(&node->link);
 	for (i = 0; i < node->n_hndl; i++) {
-		for (j = 0; j < node->desc->n_output; j++)
-			free(node->output_port[j].audio_data[i]);
 		if (node->hndl[i] == NULL)
 			continue;
 		if (d->deactivate)
 			d->deactivate(node->hndl[i]);
 		d->cleanup(node->hndl[i]);
+		node->hndl[i] = NULL;
 	}
+}
+
+static int port_ensure_data(struct port *port, uint32_t i)
+{
+	float *data;
+	if ((data = port->audio_data[i]) == NULL) {
+		data = calloc(1, MAX_SAMPLES * sizeof(float));
+		if (data == NULL) {
+			pw_log_error("cannot create port data: %m");
+			return -errno;
+		}
+	}
+	port->audio_data[i] = data;
+	return 0;
+}
+
+static void port_free_data(struct port *port, uint32_t i)
+{
+	free(port->audio_data[i]);
+	port->audio_data[i] = NULL;
+}
+
+static void node_free(struct node *node)
+{
+	uint32_t i, j;
+
+	spa_list_remove(&node->link);
+	for (i = 0; i < node->n_hndl; i++) {
+		for (j = 0; j < node->desc->n_output; j++)
+			port_free_data(&node->output_port[j], i);
+	}
+	node_cleanup(node);
 	descriptor_unref(node->desc);
 	free(node->input_port);
 	free(node->output_port);
@@ -1582,6 +1675,82 @@ static void node_free(struct node *node)
 	free(node);
 }
 
+static void graph_cleanup(struct graph *graph)
+{
+	struct node *node;
+	spa_list_for_each(node, &graph->node_list, link)
+		node_cleanup(node);
+}
+
+static int graph_instantiate(struct graph *graph)
+{
+	struct impl *impl = graph->impl;
+	struct node *node;
+	struct port *port;
+	struct link *link;
+	struct descriptor *desc;
+	const struct fc_descriptor *d;
+	uint32_t i, j;
+	int res;
+
+	spa_list_for_each(node, &graph->node_list, link) {
+		float *sd = silence_data, *dd = discard_data;
+
+		node_cleanup(node);
+
+		desc = node->desc;
+		d = desc->desc;
+		if (d->flags & FC_DESCRIPTOR_SUPPORTS_NULL_DATA)
+			sd = dd = NULL;
+
+		for (i = 0; i < node->n_hndl; i++) {
+			pw_log_info("instantiate %s %d rate:%lu", d->name, i, impl->rate);
+			if ((node->hndl[i] = d->instantiate(d, impl->rate, i, node->config)) == NULL) {
+				pw_log_error("cannot create plugin instance: %m");
+				res = -errno;
+				goto error;
+			}
+			for (j = 0; j < desc->n_input; j++) {
+				port = &node->input_port[j];
+				d->connect_port(node->hndl[i], port->p, sd);
+
+				spa_list_for_each(link, &port->link_list, input_link) {
+					struct port *peer = link->output;
+					if ((res = port_ensure_data(peer, i)) < 0)
+						goto error;
+					pw_log_info("connect input port %s[%d]:%s %p",
+							node->name, i, d->ports[port->p].name,
+							peer->audio_data[i]);
+					d->connect_port(node->hndl[i], port->p, peer->audio_data[i]);
+				}
+			}
+			for (j = 0; j < desc->n_output; j++) {
+				port = &node->output_port[j];
+				if ((res = port_ensure_data(port, i)) < 0)
+					goto error;
+				pw_log_info("connect output port %s[%d]:%s %p",
+						node->name, i, d->ports[port->p].name,
+						port->audio_data[i]);
+				d->connect_port(node->hndl[i], port->p, port->audio_data[i]);
+			}
+			for (j = 0; j < desc->n_control; j++) {
+				port = &node->control_port[j];
+				d->connect_port(node->hndl[i], port->p, &port->control_data);
+			}
+			for (j = 0; j < desc->n_notify; j++) {
+				port = &node->notify_port[j];
+				d->connect_port(node->hndl[i], port->p, &port->control_data);
+			}
+			if (d->activate)
+				d->activate(node->hndl[i]);
+		}
+	}
+	return 0;
+error:
+	graph_cleanup(graph);
+	return res;
+}
+
 static struct node *find_next_node(struct graph *graph)
 {
 	struct node *node;
@@ -1594,61 +1763,16 @@ static struct node *find_next_node(struct graph *graph)
 	return NULL;
 }
 
-static int setup_input_port(struct graph *graph, struct port *port)
-{
-	struct descriptor *desc = port->node->desc;
-	const struct fc_descriptor *d = desc->desc;
-	struct link *link;
-	uint32_t i, n_hndl = port->node->n_hndl;
-
-	spa_list_for_each(link, &port->link_list, input_link) {
-		struct port *peer = link->output;
-		for (i = 0; i < n_hndl; i++) {
-			pw_log_info("connect input port %s[%d]:%s %p",
-					port->node->name, i, d->ports[port->p].name,
-					peer->audio_data[i]);
-			d->connect_port(port->node->hndl[i], port->p, peer->audio_data[i]);
-		}
-	}
-	return 0;
-}
-
-static int setup_output_port(struct graph *graph, struct port *port)
-{
-	struct descriptor *desc = port->node->desc;
-	const struct fc_descriptor *d = desc->desc;
-	struct link *link;
-	uint32_t i, n_hndl = port->node->n_hndl;
-
-	spa_list_for_each(link, &port->link_list, output_link) {
-		for (i = 0; i < n_hndl; i++) {
-			float *data;
-			if ((data = port->audio_data[i]) == NULL) {
-				data = calloc(1, MAX_SAMPLES * sizeof(float));
-				if (data == NULL)
-					return -errno;
-			}
-			port->audio_data[i] = data;
-			pw_log_info("connect output port %s[%d]:%s %p",
-					port->node->name, i, d->ports[port->p].name,
-					port->audio_data[i]);
-			d->connect_port(port->node->hndl[i], port->p, data);
-		}
-		link->input->node->n_deps--;
-	}
-	return 0;
-}
-
 static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_json *outputs)
 {
 	struct impl *impl = graph->impl;
 	struct node *node, *first, *last;
 	struct port *port;
+	struct link *link;
 	struct graph_port *gp;
 	struct graph_hndl *gh;
 	uint32_t i, j, n_nodes, n_input, n_output, n_control, n_hndl = 0;
 	int res;
-	unsigned long p;
 	struct descriptor *desc;
 	const struct fc_descriptor *d;
 	char v[256];
@@ -1719,52 +1843,11 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 	n_control = 0;
 	n_nodes = 0;
 	spa_list_for_each(node, &graph->node_list, link) {
-		float *sd = silence_data, *dd = discard_data;
-
+		node->n_hndl = n_hndl;
 		desc = node->desc;
-		d = desc->desc;
-		if (d->flags & FC_DESCRIPTOR_SUPPORTS_NULL_DATA)
-			sd = dd = NULL;
-
-		for (i = 0; i < n_hndl; i++) {
-			pw_log_info("instantiate %s %d", d->name, i);
-			if ((node->hndl[i] = d->instantiate(d, &impl->rate, i, node->config)) == NULL) {
-				pw_log_error("cannot create plugin instance: %m");
-				res = -errno;
-				goto error;
-			}
-			node->n_hndl = i + 1;
-
-			for (j = 0; j < desc->n_input; j++) {
-				p = desc->input[j];
-				d->connect_port(node->hndl[i], p, sd);
-			}
-			for (j = 0; j < desc->n_output; j++) {
-				p = desc->output[j];
-				d->connect_port(node->hndl[i], p, dd);
-			}
-			for (j = 0; j < desc->n_control; j++) {
-				port = &node->control_port[j];
-				d->connect_port(node->hndl[i], port->p, &port->control_data);
-			}
-			for (j = 0; j < desc->n_notify; j++) {
-				port = &node->notify_port[j];
-				d->connect_port(node->hndl[i], port->p, &port->control_data);
-			}
-			if (d->activate)
-				d->activate(node->hndl[i]);
-		}
 		n_control += desc->n_control;
 		n_nodes++;
 	}
-	pw_log_info("suggested rate:%lu capture:%d playback:%d", impl->rate,
-			impl->capture_info.rate, impl->playback_info.rate);
-
-	if (impl->capture_info.rate == 0)
-		impl->capture_info.rate = impl->rate;
-	if (impl->playback_info.rate == 0)
-		impl->playback_info.rate = impl->rate;
-
 	graph->n_input = 0;
 	graph->input = calloc(n_input * n_hndl, sizeof(struct graph_port));
 	graph->n_output = 0;
@@ -1780,7 +1863,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 				pw_log_info("input port %s[%d]:%s",
 						first->name, i, d->ports[desc->input[j]].name);
 				gp->desc = d;
-				gp->hndl = first->hndl[i];
+				gp->hndl = &first->hndl[i];
 				gp->port = desc->input[j];
 			}
 		} else {
@@ -1814,7 +1897,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 							port->node->name, i, d->ports[port->p].name);
 					port->external = graph->n_input;
 					gp->desc = d;
-					gp->hndl = port->node->hndl[i];
+					gp->hndl = &port->node->hndl[i];
 					gp->port = port->p;
 				}
 				graph->n_input++;
@@ -1828,7 +1911,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 				pw_log_info("output port %s[%d]:%s",
 						last->name, i, d->ports[desc->output[j]].name);
 				gp->desc = d;
-				gp->hndl = last->hndl[i];
+				gp->hndl = &last->hndl[i];
 				gp->port = desc->output[j];
 			}
 		} else {
@@ -1862,7 +1945,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 							port->node->name, i, d->ports[port->p].name);
 					port->external = graph->n_output;
 					gp->desc = d;
-					gp->hndl = port->node->hndl[i];
+					gp->hndl = &port->node->hndl[i];
 					gp->port = port->p;
 				}
 				graph->n_output++;
@@ -1882,17 +1965,16 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 		desc = node->desc;
 		d = desc->desc;
 
-		for (i = 0; i < desc->n_input; i++)
-			setup_input_port(graph, &node->input_port[i]);
-
 		for (i = 0; i < n_hndl; i++) {
 			gh = &graph->hndl[graph->n_hndl++];
-			gh->hndl = node->hndl[i];
+			gh->hndl = &node->hndl[i];
 			gh->desc = d;
-		}
 
-		for (i = 0; i < desc->n_output; i++)
-			setup_output_port(graph, &node->output_port[i]);
+		}
+		for (i = 0; i < desc->n_output; i++) {
+			spa_list_for_each(link, &node->output_port[i].link_list, output_link)
+				link->input->node->n_deps--;
+		}
 
 		/* collect all control ports on the graph */
 		for (i = 0; i < desc->n_control; i++) {
@@ -1900,17 +1982,8 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 			graph->n_control++;
 		}
 	}
-	return 0;
-
+	res = 0;
 error:
-	spa_list_for_each(node, &graph->node_list, link) {
-		for (i = 0; i < node->n_hndl; i++) {
-			if (node->hndl[i] != NULL)
-				node->desc->desc->cleanup(node->hndl[i]);
-			node->hndl[i] = NULL;
-		}
-		node->n_hndl = 0;
-	}
 	return res;
 }
 
@@ -2041,12 +2114,20 @@ static const struct pw_proxy_events core_proxy_events = {
 
 static void impl_destroy(struct impl *impl)
 {
+	/* disconnect both streams before destroying any of them */
+	if (impl->capture)
+		pw_stream_disconnect(impl->capture);
+	if (impl->playback)
+		pw_stream_disconnect(impl->playback);
+
 	if (impl->capture)
 		pw_stream_destroy(impl->capture);
 	if (impl->playback)
 		pw_stream_destroy(impl->playback);
+
 	if (impl->core && impl->do_disconnect)
 		pw_core_disconnect(impl->core);
+
 	pw_properties_free(impl->capture_props);
 	pw_properties_free(impl->playback_props);
 	graph_free(&impl->graph);
@@ -2156,7 +2237,6 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	impl->module = module;
 	impl->context = context;
 
-	impl->rate = 48000;
 	impl->graph.impl = impl;
 	spa_list_init(&impl->plugin_list);
 
@@ -2166,6 +2246,8 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 		pw_properties_setf(props, PW_KEY_NODE_LINK_GROUP, "filter-chain-%u-%u", pid, id);
 	if (pw_properties_get(props, PW_KEY_NODE_VIRTUAL) == NULL)
 		pw_properties_set(props, PW_KEY_NODE_VIRTUAL, "true");
+	if (pw_properties_get(props, "resample.prefill") == NULL)
+		pw_properties_set(props, "resample.prefill", "true");
 	if (pw_properties_get(props, PW_KEY_NODE_DESCRIPTION) == NULL)
 		pw_properties_setf(props, PW_KEY_NODE_DESCRIPTION, "filter-chain-%u-%u", pid, id);
 
@@ -2183,6 +2265,7 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	copy_props(impl, props, PW_KEY_NODE_LATENCY);
 	copy_props(impl, props, PW_KEY_NODE_VIRTUAL);
 	copy_props(impl, props, PW_KEY_MEDIA_NAME);
+	copy_props(impl, props, "resample.prefill");
 
 	parse_audio_info(impl->capture_props, &impl->capture_info);
 	parse_audio_info(impl->playback_props, &impl->playback_info);
