@@ -31,7 +31,6 @@
 
 #include "../defs.h"
 #include "../module.h"
-#include "registry.h"
 
 #define NAME "loopback"
 
@@ -48,6 +47,7 @@ struct module_loopback_data {
 	struct pw_properties *playback_props;
 
 	struct spa_audio_info_raw info;
+	uint32_t latency_msec;
 };
 
 static void module_destroy(void *data)
@@ -69,11 +69,16 @@ static int module_loopback_load(struct client *client, struct module *module)
 	FILE *f;
 	char *args;
 	size_t size, i;
+	char val[256];
 
 	pw_properties_setf(data->capture_props, PW_KEY_NODE_GROUP, "loopback-%u", module->index);
 	pw_properties_setf(data->playback_props, PW_KEY_NODE_GROUP, "loopback-%u", module->index);
+	pw_properties_setf(data->capture_props, "pulse.module.id", "%u", module->index);
+	pw_properties_setf(data->playback_props, "pulse.module.id", "%u", module->index);
+
+	if ((f = open_memstream(&args, &size)) == NULL)
+		return -errno;
 
-	f = open_memstream(&args, &size);
 	fprintf(f, "{");
 	if (data->info.channels != 0) {
 		fprintf(f, " audio.channels = %u", data->info.channels);
@@ -85,6 +90,10 @@ static int module_loopback_load(struct client *client, struct module *module)
 			fprintf(f, " ]");
 		}
 	}
+	if (data->latency_msec != 0)
+		fprintf(f, " target.delay.sec = %s",
+				spa_json_format_float(val, sizeof(val),
+					data->latency_msec / 1000.0f));
 	fprintf(f, " capture.props = {");
 	pw_properties_serialize_dict(f, &data->capture_props->dict, 0);
 	fprintf(f, " } playback.props = {");
@@ -123,12 +132,6 @@ static int module_loopback_unload(struct module *module)
 	return 0;
 }
 
-static const struct module_methods module_loopback_methods = {
-	VERSION_MODULE_METHODS,
-	.load = module_loopback_load,
-	.unload = module_loopback_unload,
-};
-
 static const struct spa_dict_item module_loopback_info[] = {
 	{ PW_KEY_MODULE_AUTHOR, "Arun Raghavan <arun@asymptotic.io>" },
 	{ PW_KEY_MODULE_DESCRIPTION, "Loopback from source to sink" },
@@ -146,26 +149,23 @@ static const struct spa_dict_item module_loopback_info[] = {
 	{ PW_KEY_MODULE_VERSION, PACKAGE_VERSION },
 };
 
-struct module *create_module_loopback(struct impl *impl, const char *argument)
+static int module_loopback_prepare(struct module * const module)
 {
-	struct module *module;
-	struct module_loopback_data *d;
-	struct pw_properties *props = NULL, *playback_props = NULL, *capture_props = NULL;
+	struct module_loopback_data * const d = module->user_data;
+	struct pw_properties * const props = module->props;
+	struct pw_properties *playback_props = NULL, *capture_props = NULL;
 	const char *str;
 	struct spa_audio_info_raw info = { 0 };
 	int res;
 
 	PW_LOG_TOPIC_INIT(mod_topic);
 
-	props = pw_properties_new_dict(&SPA_DICT_INIT_ARRAY(module_loopback_info));
 	capture_props = pw_properties_new(NULL, NULL);
 	playback_props = pw_properties_new(NULL, NULL);
-	if (!props || !capture_props || !playback_props) {
+	if (!capture_props || !playback_props) {
 		res = -EINVAL;
 		goto out;
 	}
-	if (argument)
-		module_args_add_props(props, argument);
 
 	/* The following modargs are not implemented:
 	 * adjust_time, max_latency_msec, fast_adjust_threshold_msec: these are just not relevant
@@ -188,7 +188,7 @@ struct module *create_module_loopback(struct impl *impl, const char *argument)
 		pw_properties_set(props, "sink", NULL);
 	}
 
-	if (module_args_to_audioinfo(impl, props, &info) < 0) {
+	if (module_args_to_audioinfo(module->impl, props, &info) < 0) {
 		res = -EINVAL;
 		goto out;
 	}
@@ -210,15 +210,8 @@ struct module *create_module_loopback(struct impl *impl, const char *argument)
 		pw_properties_set(props, "remix", NULL);
 	}
 
-	if ((str = pw_properties_get(props, "latency_msec")) != NULL) {
-		/* Half the latency on each of the playback and capture streams */
-		pw_properties_setf(capture_props, PW_KEY_NODE_LATENCY, "%s/2000", str);
-		pw_properties_setf(playback_props, PW_KEY_NODE_LATENCY, "%s/2000", str);
-		pw_properties_set(props, "latency_msec", NULL);
-	} else {
-		pw_properties_set(capture_props, PW_KEY_NODE_LATENCY, "100/1000");
-		pw_properties_set(playback_props, PW_KEY_NODE_LATENCY, "100/1000");
-	}
+	if ((str = pw_properties_get(props, "latency_msec")) != NULL)
+		d->latency_msec = atoi(str);
 
 	if ((str = pw_properties_get(props, "sink_input_properties")) != NULL) {
 		module_args_add_props(playback_props, str);
@@ -230,25 +223,24 @@ struct module *create_module_loopback(struct impl *impl, const char *argument)
 		pw_properties_set(props, "source_output_properties", NULL);
 	}
 
-	module = module_new(impl, &module_loopback_methods, sizeof(*d));
-	if (module == NULL) {
-		res = -errno;
-		goto out;
-	}
-
-	module->props = props;
-	d = module->user_data;
 	d->module = module;
 	d->capture_props = capture_props;
 	d->playback_props = playback_props;
 	d->info = info;
 
-	return module;
+	return 0;
 out:
-	pw_properties_free(props);
 	pw_properties_free(playback_props);
 	pw_properties_free(capture_props);
-	errno = -res;
 
-	return NULL;
+	return res;
 }
+
+DEFINE_MODULE_INFO(module_loopback) = {
+	.name = "module-loopback",
+	.prepare = module_loopback_prepare,
+	.load = module_loopback_load,
+	.unload = module_loopback_unload,
+	.properties = &SPA_DICT_INIT_ARRAY(module_loopback_info),
+	.data_size = sizeof(struct module_loopback_data),
+};
