@@ -166,6 +166,8 @@ static int handle_memblock(struct client *client, struct message *msg)
 	index += diff;
 	filled += diff;
 	stream->write_index += diff;
+	if ((flags & FLAG_SEEKMASK) == SEEK_RELATIVE)
+		stream->requested -= diff;
 
 	if (filled < 0) {
 		/* underrun, reported on reader side */
@@ -182,12 +184,16 @@ static int handle_memblock(struct client *client, struct message *msg)
 			msg->data,
 			SPA_MIN(msg->length, MAXLENGTH));
 	index += msg->length;
-	stream->write_index += msg->length;
 	spa_ringbuffer_write_update(&stream->ring, index);
-	stream->requested -= SPA_MIN(msg->length, stream->requested);
+
+	stream->write_index += msg->length;
+	stream->requested -= msg->length;
 
 	stream_send_request(stream);
 
+	if (stream->is_paused && !stream->corked)
+		stream_set_paused(stream, false, "new data");
+
 finish:
 	message_free(msg, false, false);
 	return res;
@@ -613,7 +619,6 @@ static int start_unix_server(struct server *server, const struct sockaddr_storag
 			pw_log_warn("server %p: unlink('%s') failed: %m",
 				    server, addr_un->sun_path);
 	}
-
 	if (bind(fd, (const struct sockaddr *) addr_un, SUN_LEN(addr_un)) < 0) {
 		res = -errno;
 		pw_log_warn("server %p: bind() to '%s' failed: %m",
@@ -621,6 +626,10 @@ static int start_unix_server(struct server *server, const struct sockaddr_storag
 		goto error_close;
 	}
 
+	if (chmod(addr_un->sun_path, 0777) < 0)
+		pw_log_warn("server %p: chmod('%s') failed: %m",
+				    server, addr_un->sun_path);
+
 	if (listen(fd, server->listen_backlog) < 0) {
 		res = -errno;
 		pw_log_warn("server %p: listen() on '%s' failed: %m",
@@ -893,7 +902,7 @@ static struct server *server_new(struct impl *impl)
 
 static int server_start(struct server *server, const struct sockaddr_storage *addr)
 {
-	const struct impl * const impl = server->impl;
+	struct impl * const impl = server->impl;
 	int res = 0, fd;
 
 	switch (addr->ss_family) {
@@ -918,6 +927,8 @@ static int server_start(struct server *server, const struct sockaddr_storage *ad
 		res = -errno;
 		pw_log_error("server %p: can't create server source: %m", impl);
 	}
+	if (res >= 0)
+		spa_hook_list_call(&impl->hooks, struct impl_events, server_started, 0, server);
 
 	return res;
 }
@@ -1063,6 +1074,8 @@ void server_free(struct server *server)
 		client_unref(c);
 	}
 
+	spa_hook_list_call(&impl->hooks, struct impl_events, server_stopped, 0, server);
+
 	if (server->source)
 		pw_loop_destroy_source(impl->loop, server->source);
 
