@@ -35,6 +35,7 @@
 #include <spa/utils/ringbuffer.h>
 #include <spa/utils/string.h>
 #include <spa/pod/filter.h>
+#include <spa/pod/dynamic.h>
 #include <spa/debug/format.h>
 #include <spa/debug/types.h>
 #include <spa/debug/pod.h>
@@ -50,9 +51,7 @@ PW_LOG_TOPIC_EXTERN(log_filter);
 #define MAX_BUFFERS	64
 
 #define MASK_BUFFERS	(MAX_BUFFERS-1)
-#define MAX_PORTS	1024
 
-static float empty[MAX_SAMPLES];
 static bool mlock_warned = false;
 
 static uint32_t mappable_dataTypes = (1<<SPA_DATA_MemFd);
@@ -143,14 +142,15 @@ struct filter {
 	} rt;
 
 	struct spa_list port_list;
-	struct port *ports[2][MAX_PORTS];
+	struct pw_map ports[2];
 
 	uint32_t change_mask_all;
 	struct spa_node_info info;
 	struct spa_list param_list;
-#define IDX_Props		0
-#define IDX_ProcessLatency	1
-#define N_NODE_PARAMS		2
+#define IDX_PropInfo		0
+#define IDX_Props		1
+#define IDX_ProcessLatency	2
+#define N_NODE_PARAMS		3
 	struct spa_param_info params[N_NODE_PARAMS];
 
 	struct spa_process_latency_info process_latency;
@@ -175,6 +175,8 @@ struct filter {
 static int get_param_index(uint32_t id)
 {
 	switch (id) {
+	case SPA_PARAM_PropInfo:
+		return IDX_PropInfo;
 	case SPA_PARAM_Props:
 		return IDX_Props;
 	case SPA_PARAM_ProcessLatency:
@@ -307,27 +309,17 @@ static struct port *alloc_port(struct filter *filter,
 		enum spa_direction direction, uint32_t user_data_size)
 {
 	struct port *p;
-	int i;
-
-	for (i = 0; i < MAX_PORTS; i++) {
-		if ((filter->ports[direction][i]) == NULL)
-			break;
-	}
-	if (i == MAX_PORTS)
-		return NULL;
 
 	p = calloc(1, sizeof(struct port) + user_data_size);
 	p->filter = filter;
 	p->direction = direction;
-	p->id = i;
 	p->latency[SPA_DIRECTION_INPUT] = SPA_LATENCY_INFO(SPA_DIRECTION_INPUT);
 	p->latency[SPA_DIRECTION_OUTPUT] = SPA_LATENCY_INFO(SPA_DIRECTION_OUTPUT);
 
 	spa_list_init(&p->param_list);
 	spa_ringbuffer_init(&p->dequeued.ring);
 	spa_ringbuffer_init(&p->queued.ring);
-
-	filter->ports[direction][i] = p;
+	p->id = pw_map_insert_new(&filter->ports[direction], p);
 	spa_list_append(&filter->port_list, &p->link);
 
 	return p;
@@ -335,10 +327,9 @@ static struct port *alloc_port(struct filter *filter,
 
 static inline struct port *get_port(struct filter *filter, enum spa_direction direction, uint32_t port_id)
 {
-	if ((direction != SPA_DIRECTION_INPUT && direction != SPA_DIRECTION_OUTPUT) ||
-	    port_id >= MAX_PORTS)
+	if ((direction != SPA_DIRECTION_INPUT && direction != SPA_DIRECTION_OUTPUT))
 		return NULL;
-	return filter->ports[direction][port_id];
+	return pw_map_lookup(&filter->ports[direction], port_id);
 }
 
 static inline int push_queue(struct port *port, struct queue *queue, struct buffer *buffer)
@@ -411,7 +402,7 @@ static int enum_params(struct filter *d, struct spa_list *param_list, int seq,
 {
 	struct spa_result_node_params result;
 	uint8_t buffer[1024];
-	struct spa_pod_builder b = { 0 };
+	struct spa_pod_dynamic_builder b;
 	uint32_t count = 0;
 	struct param *p;
 	bool found = false;
@@ -428,23 +419,24 @@ static int enum_params(struct filter *d, struct spa_list *param_list, int seq,
 	spa_list_for_each(p, param_list, link) {
 		struct spa_pod *param;
 
-		result.index = result.next++;
-		if (result.index < start)
-			continue;
-
 		param = p->param;
 		if (param == NULL || p->id != id)
 			continue;
 
 		found = true;
 
-		spa_pod_builder_init(&b, buffer, sizeof(buffer));
-		if (spa_pod_filter(&b, &result.param, param, filter) != 0)
+		result.index = result.next++;
+		if (result.index < start)
 			continue;
 
-		spa_node_emit_result(&d->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
+		spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
+		if (spa_pod_filter(&b.b, &result.param, param, filter) == 0) {
+			spa_node_emit_result(&d->hooks, seq, 0, SPA_RESULT_TYPE_NODE_PARAMS, &result);
+			count++;
+		}
+		spa_pod_dynamic_builder_clean(&b);
 
-		if (++count == num)
+		if (count == num)
 			break;
 	}
 	return found ? 0 : -ENOENT;
@@ -650,7 +642,7 @@ static int update_params(struct filter *impl, struct port *port, uint32_t id,
 		clear_params(impl, port, id);
 	} else {
 		for (i = 0; i < n_params; i++) {
-			if (!spa_pod_is_object(params[i]))
+			if (params[i] == NULL || !spa_pod_is_object(params[i]))
 				continue;
 			clear_params(impl, port, SPA_POD_OBJECT_ID(params[i]));
 		}
@@ -982,7 +974,7 @@ static inline void copy_position(struct filter *impl)
 			impl->clock_id = p->clock.id;
 		}
 		impl->time.ticks = p->clock.position - impl->base_pos;
-		impl->time.delay = p->clock.delay;
+		impl->time.delay = 0;
 		SEQ_WRITE(impl->seq);
 	}
 }
@@ -1183,6 +1175,22 @@ static const struct pw_core_events core_events = {
 	.error = on_core_error,
 };
 
+struct match {
+	struct pw_filter *filter;
+	int count;
+};
+#define MATCH_INIT(f) ((struct match){ .filter = (f) })
+
+static int execute_match(void *data, const char *location, const char *action,
+		const char *val, size_t len)
+{
+	struct match *match = data;
+	struct pw_filter *this = match->filter;
+	if (spa_streq(action, "update-props"))
+		match->count += pw_properties_update_string(this->properties, val, len);
+	return 1;
+}
+
 static struct filter *
 filter_new(struct pw_context *context, const char *name,
 		struct pw_properties *props, const struct pw_properties *extra)
@@ -1190,6 +1198,7 @@ filter_new(struct pw_context *context, const char *name,
 	struct filter *impl;
 	struct pw_filter *this;
 	const char *str;
+	struct match match;
 	int res;
 
 	impl = calloc(1, sizeof(struct filter));
@@ -1210,8 +1219,30 @@ filter_new(struct pw_context *context, const char *name,
 		res = -errno;
 		goto error_properties;
 	}
-	if ((str = pw_context_get_conf_section(context, "filter.properties")) != NULL)
+	spa_hook_list_init(&impl->hooks);
+	this->properties = props;
+
+	pw_context_conf_update_props(context, "filter.properties", props);
+
+	match = MATCH_INIT(this);
+	pw_context_conf_section_match_rules(context, "filter.rules",
+		&this->properties->dict, execute_match, &match);
+
+	if ((str = getenv("PIPEWIRE_PROPS")) != NULL)
 		pw_properties_update_string(props, str, strlen(str));
+	if ((str = getenv("PIPEWIRE_QUANTUM")) != NULL) {
+		struct spa_fraction q;
+		if (sscanf(str, "%u/%u", &q.num, &q.denom) == 2 && q.denom != 0) {
+			pw_properties_setf(props, PW_KEY_NODE_RATE,
+					"1/%u", q.denom);
+			pw_properties_setf(props, PW_KEY_NODE_LATENCY,
+					"%u/%u", q.num, q.denom);
+		}
+	}
+	if ((str = getenv("PIPEWIRE_LATENCY")) != NULL)
+		pw_properties_set(props, PW_KEY_NODE_LATENCY, str);
+	if ((str = getenv("PIPEWIRE_RATE")) != NULL)
+		pw_properties_set(props, PW_KEY_NODE_RATE, str);
 
 	if (pw_properties_get(props, PW_KEY_NODE_NAME) == NULL && extra) {
 		str = pw_properties_get(extra, PW_KEY_APP_NAME);
@@ -1221,17 +1252,14 @@ filter_new(struct pw_context *context, const char *name,
 			str = name;
 		pw_properties_set(props, PW_KEY_NODE_NAME, str);
 	}
-	if ((str = getenv("PIPEWIRE_LATENCY")) != NULL)
-		pw_properties_set(props, PW_KEY_NODE_LATENCY, str);
-
-	spa_hook_list_init(&impl->hooks);
-	this->properties = props;
 
 	this->name = name ? strdup(name) : NULL;
 	this->node_id = SPA_ID_INVALID;
 
 	spa_list_init(&impl->param_list);
 	spa_list_init(&impl->port_list);
+	pw_map_init(&impl->ports[SPA_DIRECTION_INPUT], 32, 32);
+	pw_map_init(&impl->ports[SPA_DIRECTION_OUTPUT], 32, 32);
 
 	spa_hook_list_init(&this->listener_list);
 	spa_list_init(&this->controls);
@@ -1368,6 +1396,9 @@ void pw_filter_destroy(struct pw_filter *filter)
 	spa_hook_list_clean(&impl->hooks);
 	spa_hook_list_clean(&filter->listener_list);
 
+	pw_map_clear(&impl->ports[SPA_DIRECTION_INPUT]);
+	pw_map_clear(&impl->ports[SPA_DIRECTION_OUTPUT]);
+
 	free(filter->name);
 
 	if (impl->data.context)
@@ -1447,9 +1478,16 @@ int pw_filter_update_properties(struct pw_filter *filter, void *port_data, const
 			emit_port_info(impl, port, false);
 		}
 	} else {
+		struct match match;
+
 		changed = pw_properties_update(filter->properties, dict);
+
+		match = MATCH_INIT(filter);
+		pw_context_conf_section_match_rules(impl->context, "filter.rules",
+			&filter->properties->dict, execute_match, &match);
+
 		impl->info.props = &filter->properties->dict;
-		if (changed > 0) {
+		if (changed > 0 || match.count > 0) {
 			impl->info.change_mask |= SPA_NODE_CHANGE_MASK_PROPS;
 			emit_node_info(impl, false);
 		}
@@ -1487,10 +1525,11 @@ pw_filter_connect(struct pw_filter *filter,
 		SPA_NODE_CHANGE_MASK_PARAMS;
 
 	impl->info = SPA_NODE_INFO_INIT();
-	impl->info.max_input_ports = MAX_PORTS;
-	impl->info.max_output_ports = MAX_PORTS;
+	impl->info.max_input_ports = UINT32_MAX;
+	impl->info.max_output_ports = UINT32_MAX;
 	impl->info.flags = impl->process_rt ? SPA_NODE_FLAG_RT : 0;
 	impl->info.props = &filter->properties->dict;
+	impl->params[IDX_PropInfo] = SPA_PARAM_INFO(SPA_PARAM_PropInfo, 0);
 	impl->params[IDX_Props] = SPA_PARAM_INFO(SPA_PARAM_Props, SPA_PARAM_INFO_WRITE);
 	impl->params[IDX_ProcessLatency] = SPA_PARAM_INFO(SPA_PARAM_ProcessLatency, 0);
 	impl->info.params = impl->params;
@@ -1505,6 +1544,11 @@ pw_filter_connect(struct pw_filter *filter,
 	impl->disconnecting = false;
 	filter_set_state(filter, PW_FILTER_STATE_CONNECTING, NULL);
 
+	if (flags & PW_FILTER_FLAG_DRIVER)
+		pw_properties_set(filter->properties, PW_KEY_NODE_DRIVER, "true");
+	if ((pw_properties_get(filter->properties, PW_KEY_NODE_WANT_DRIVER) == NULL))
+		pw_properties_set(filter->properties, PW_KEY_NODE_WANT_DRIVER, "true");
+
 	if (filter->core == NULL) {
 		filter->core = pw_context_connect(impl->context,
 				pw_properties_copy(filter->properties), 0);
@@ -1606,8 +1650,7 @@ static void add_audio_dsp_port_params(struct filter *impl, struct port *port)
 								sizeof(float),
 								MAX_SAMPLES * sizeof(float),
 								sizeof(float)),
-			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(4),
-			SPA_PARAM_BUFFERS_align,   SPA_POD_Int(16)));
+			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(4)));
 }
 
 static void add_video_dsp_port_params(struct filter *impl, struct port *port)
@@ -1714,7 +1757,7 @@ int pw_filter_remove_port(void *port_data)
 	spa_node_emit_port_info(&impl->hooks, port->direction, port->id, NULL);
 
 	spa_list_remove(&port->link);
-	impl->ports[port->direction][port->id] = NULL;
+	pw_map_remove(&impl->ports[port->direction], port->id);
 
 	clear_buffers(port);
 	clear_params(impl, port, SPA_ID_INVALID);
@@ -1861,7 +1904,7 @@ void *pw_filter_get_dsp_buffer(void *port_data, uint32_t n_samples)
 	struct spa_data *d;
 
 	if ((buf = pw_filter_dequeue_buffer(port_data)) == NULL)
-		return empty;
+		return NULL;
 
 	d = &buf->buffer->datas[0];
 
