@@ -281,8 +281,10 @@ static int do_negotiate(struct pw_impl_link *this)
 	/* find a common format for the ports */
 	if ((res = pw_context_find_format(context,
 					output, input, NULL, 0, NULL,
-					&format, &b, &error)) < 0)
+					&format, &b, &error)) < 0) {
+		format = NULL;
 		goto error;
+	}
 
 	format = spa_pod_copy(format);
 	spa_pod_fixate(format);
@@ -358,10 +360,11 @@ static int do_negotiate(struct pw_impl_link *this)
 		}
 	}
 
-	pw_log_debug("%p: doing set format %p", this, format);
 	pw_log_pod(SPA_LOG_LEVEL_DEBUG, format);
 
 	SPA_POD_OBJECT_ID(format) = SPA_PARAM_Format;
+	pw_log_debug("%p: doing set format %p fixated:%d", this,
+			format, spa_pod_is_fixated(format));
 
 	if (out_state == PW_IMPL_PORT_STATE_CONFIGURE) {
 		pw_log_debug("%p: doing set format on output", this);
@@ -450,7 +453,7 @@ static int port_set_io(struct pw_impl_link *this, struct pw_impl_port *port, uin
 	return res;
 }
 
-static int select_io(struct pw_impl_link *this)
+static void select_io(struct pw_impl_link *this)
 {
 	struct impl *impl = SPA_CONTAINER_OF(this, struct impl, this);
 	struct spa_io_buffers *io;
@@ -460,13 +463,9 @@ static int select_io(struct pw_impl_link *this)
 		io = this->rt.out_mix.io;
 	if (io == NULL)
 		io = &impl->io;
-	if (io == NULL)
-		return -EIO;
 
 	this->io = io;
 	*this->io = SPA_IO_BUFFERS_INIT;
-
-	return 0;
 }
 
 static int do_allocation(struct pw_impl_link *this)
@@ -485,6 +484,9 @@ static int do_allocation(struct pw_impl_link *this)
 
 	pw_log_debug("%p: out-state:%d in-state:%d", this, output->state, input->state);
 
+	if (input->state < PW_IMPL_PORT_STATE_READY || output->state < PW_IMPL_PORT_STATE_READY)
+		return 0;
+
 	link_update_state(this, PW_LINK_STATE_ALLOCATING, 0, NULL);
 
 	out_flags = output->spa_flags;
@@ -512,6 +514,12 @@ static int do_allocation(struct pw_impl_link *this)
 		flags = 0;
 		/* always shared buffers for the link */
 		alloc_flags = PW_BUFFERS_FLAG_SHARED;
+		if (output->node->remote || input->node->remote)
+			alloc_flags |= PW_BUFFERS_FLAG_SHARED_MEM;
+
+		if (output->node->driver)
+			alloc_flags |= PW_BUFFERS_FLAG_IN_PRIORITY;
+
 		/* if output port can alloc buffers, alloc skeleton buffers */
 		if (SPA_FLAG_IS_SET(out_flags, SPA_PORT_FLAG_CAN_ALLOC_BUFFERS)) {
 			SPA_FLAG_SET(alloc_flags, PW_BUFFERS_FLAG_NO_MEM);
@@ -614,7 +622,8 @@ int pw_impl_link_activate(struct pw_impl_link *this)
 	pw_log_debug("%p: activate activated:%d state:%s", this, impl->activated,
 			pw_link_state_as_string(this->info.state));
 
-	if (impl->activated || !this->prepared || !impl->inode->active || !impl->onode->active)
+	if (impl->activated || !this->prepared || !impl->inode->active ||
+			!impl->inode->added || !impl->onode->active)
 		return 0;
 
 	if (!impl->io_set) {
@@ -846,10 +855,10 @@ int pw_impl_link_deactivate(struct pw_impl_link *this)
 }
 
 static int
-global_bind(void *_data, struct pw_impl_client *client, uint32_t permissions,
+global_bind(void *object, struct pw_impl_client *client, uint32_t permissions,
 	       uint32_t version, uint32_t id)
 {
-	struct pw_impl_link *this = _data;
+	struct pw_impl_link *this = object;
 	struct pw_global *global = this->global;
 	struct pw_resource *resource;
 
@@ -912,8 +921,6 @@ static void port_param_changed(struct pw_impl_link *this, uint32_t id,
 	case SPA_PARAM_EnumFormat:
 		target = PW_IMPL_PORT_STATE_CONFIGURE;
 		break;
-	case SPA_PARAM_Latency:
-		return;
 	default:
 		return;
 	}
@@ -922,6 +929,8 @@ static void port_param_changed(struct pw_impl_link *this, uint32_t id,
 	if (inport)
 		pw_impl_port_update_state(inport, target, 0, NULL);
 
+	this->preparing = this->prepared = false;
+	link_update_state(this, PW_LINK_STATE_INIT, 0, NULL);
 	pw_impl_link_prepare(this);
 }
 
@@ -1126,7 +1135,7 @@ static void permissions_changed(struct pw_impl_link *this, struct pw_impl_port *
 
 	if (check_permission(this->context, this->output, this->input, this->properties) < 0) {
 		pw_impl_link_destroy(this);
-	} else {
+	} else if (this->global != NULL) {
 		pw_global_update_permissions(this->global, client, old, new);
 	}
 }
@@ -1206,8 +1215,6 @@ struct pw_impl_link *pw_context_create_link(struct pw_context *context,
                 this->user_data = SPA_PTROFF(impl, sizeof(struct impl), void);
 
 	impl->work = pw_context_get_work_queue(context);
-	if (impl->work == NULL)
-		goto error_work_queue;
 
 	this->context = context;
 	this->properties = properties;
@@ -1222,6 +1229,16 @@ struct pw_impl_link *pw_context_create_link(struct pw_context *context,
 	spa_hook_list_init(&this->listener_list);
 
 	impl->format_filter = format_filter;
+	this->info.format = NULL;
+	this->info.props = &this->properties->dict;
+
+	this->rt.out_mix.peer_id = input->global->id;
+	this->rt.in_mix.peer_id = output->global->id;
+
+	if ((res = pw_impl_port_init_mix(output, &this->rt.out_mix)) < 0)
+		goto error_output_mix;
+	if ((res = pw_impl_port_init_mix(input, &this->rt.in_mix)) < 0)
+		goto error_input_mix;
 
 	pw_impl_port_add_listener(input, &impl->input_port_listener, &input_port_events, impl);
 	pw_impl_node_add_listener(input_node, &impl->input_node_listener, &input_node_events, impl);
@@ -1238,19 +1255,9 @@ struct pw_impl_link *pw_context_create_link(struct pw_context *context,
 	spa_list_append(&output->links, &this->output_link);
 	spa_list_append(&input->links, &this->input_link);
 
-	this->info.format = NULL;
-	this->info.props = &this->properties->dict;
-
 	impl->io = SPA_IO_BUFFERS_INIT;
 
-	this->rt.out_mix.peer_id = input->global->id;
-	this->rt.in_mix.peer_id = output->global->id;
-
-	pw_impl_port_init_mix(output, &this->rt.out_mix);
-	pw_impl_port_init_mix(input, &this->rt.in_mix);
-
-	if ((res = select_io(this)) < 0)
-		goto error_no_io;
+	select_io(this);
 
 	if (this->feedback) {
 		impl->inode = output_node;
@@ -1261,7 +1268,7 @@ struct pw_impl_link *pw_context_create_link(struct pw_context *context,
 		impl->inode = input_node;
 	}
 
-	this->rt.target.signal = impl->inode->rt.target.signal;
+	this->rt.target.signal_func = impl->inode->rt.target.signal_func;
 	this->rt.target.data = impl->inode->rt.target.data;
 
 	pw_log_debug("%p: constructed out:%p:%d.%d -> in:%p:%d.%d", impl,
@@ -1279,8 +1286,8 @@ struct pw_impl_link *pw_context_create_link(struct pw_context *context,
 
 	try_link_controls(impl, output, input);
 
-	pw_impl_port_recalc_latency(this->output);
-	pw_impl_port_recalc_latency(this->input);
+	pw_impl_port_recalc_latency(output);
+	pw_impl_port_recalc_latency(input);
 
 	pw_impl_node_emit_peer_added(impl->onode, impl->inode);
 
@@ -1306,12 +1313,12 @@ error_no_mem:
 	res = -errno;
 	pw_log_debug("alloc failed: %m");
 	goto error_exit;
-error_work_queue:
-	res = -errno;
-	pw_log_debug("work queue failed: %m");
+error_output_mix:
+	pw_log_error("%p: can't get output mix %d (%s)", this, res, spa_strerror(res));
 	goto error_free;
-error_no_io:
-	pw_log_debug("%p: can't set io %d (%s)", this, res, spa_strerror(res));
+error_input_mix:
+	pw_log_error("%p: can't get input mix %d (%s)", this, res, spa_strerror(res));
+	pw_impl_port_release_mix(output, &this->rt.out_mix);
 	goto error_free;
 error_free:
 	free(impl);
@@ -1321,9 +1328,9 @@ error_exit:
 	return NULL;
 }
 
-static void global_destroy(void *object)
+static void global_destroy(void *data)
 {
-	struct pw_impl_link *link = object;
+	struct pw_impl_link *link = data;
 	spa_hook_remove(&link->global_listener);
 	link->global = NULL;
 	pw_impl_link_destroy(link);
@@ -1339,6 +1346,7 @@ int pw_impl_link_register(struct pw_impl_link *link,
 		     struct pw_properties *properties)
 {
 	static const char * const keys[] = {
+		PW_KEY_OBJECT_SERIAL,
 		PW_KEY_OBJECT_PATH,
 		PW_KEY_MODULE_ID,
 		PW_KEY_FACTORY_ID,
@@ -1378,6 +1386,8 @@ int pw_impl_link_register(struct pw_impl_link *link,
 
 	link->info.id = link->global->id;
 	pw_properties_setf(link->properties, PW_KEY_OBJECT_ID, "%d", link->info.id);
+	pw_properties_setf(link->properties, PW_KEY_OBJECT_SERIAL, "%"PRIu64,
+			pw_global_get_serial(link->global));
 	pw_properties_setf(link->properties, PW_KEY_LINK_OUTPUT_NODE, "%u", link->info.output_node_id);
 	pw_properties_setf(link->properties, PW_KEY_LINK_OUTPUT_PORT, "%u", link->info.output_port_id);
 	pw_properties_setf(link->properties, PW_KEY_LINK_INPUT_NODE, "%u", link->info.input_node_id);
