@@ -49,7 +49,7 @@ struct ucred {
 #define spa_debug(...) pw_log_trace(__VA_ARGS__)
 #endif
 
-#define MAX_RATES				16u
+#define MAX_RATES				32u
 #define CLOCK_MIN_QUANTUM			4u
 #define CLOCK_MAX_QUANTUM			65536u
 
@@ -104,6 +104,7 @@ static inline bool ratelimit_test(struct ratelimit *r, uint64_t now, enum spa_lo
 
 struct pw_param {
 	uint32_t id;
+	int32_t seq;
 	struct spa_list link;
 	struct spa_pod *param;
 };
@@ -123,7 +124,7 @@ static inline uint32_t pw_param_clear(struct spa_list *param_list, uint32_t id)
 	return count;
 }
 
-static inline struct pw_param *pw_param_add(struct spa_list *params,
+static inline struct pw_param *pw_param_add(struct spa_list *params, int32_t seq,
 		uint32_t id, const struct spa_pod *param)
 {
 	struct pw_param *p;
@@ -140,6 +141,7 @@ static inline struct pw_param *pw_param_add(struct spa_list *params,
 		return NULL;
 
 	p->id = id;
+	p->seq = seq;
 	if (param != NULL) {
 		p->param = SPA_PTROFF(p, sizeof(*p), struct spa_pod);
 		memcpy(p->param, param, SPA_POD_SIZE(param));
@@ -151,10 +153,22 @@ static inline struct pw_param *pw_param_add(struct spa_list *params,
 	return p;
 }
 
-static inline void pw_param_update(struct spa_list *param_list, struct spa_list *pending_list)
+static inline void pw_param_update(struct spa_list *param_list, struct spa_list *pending_list,
+			uint32_t n_params, struct spa_param_info *params)
 {
-	struct pw_param *p;
+	struct pw_param *p, *t;
+	uint32_t i;
 
+	for (i = 0; i < n_params; i++) {
+		spa_list_for_each_safe(p, t, pending_list, link) {
+			if (p->id == params[i].id &&
+			    p->seq != params[i].seq &&
+			    p->param != NULL) {
+				spa_list_remove(&p->link);
+				free(p);
+			}
+		}
+	}
 	spa_list_consume(p, pending_list, link) {
 		spa_list_remove(&p->link);
 		if (p->param == NULL) {
@@ -177,7 +191,7 @@ static inline struct spa_param_info *pw_param_info_find(struct spa_param_info in
 	return NULL;
 }
 
-#define pw_protocol_emit_destroy(p) spa_hook_list_call(&p->listener_list, struct pw_protocol_events, destroy, 0)
+#define pw_protocol_emit_destroy(p) spa_hook_list_call(&(p)->listener_list, struct pw_protocol_events, destroy, 0)
 
 struct pw_protocol {
 	struct spa_list link;                   /**< link in context protocol_list */
@@ -704,6 +718,7 @@ struct pw_impl_node {
 	unsigned int transport_sync:1;	/**< supports transport sync */
 	unsigned int current_pending:1;	/**< a quantum/rate update is pending */
 	unsigned int moved:1;		/**< the node was moved drivers */
+	unsigned int added:1;		/**< the node was add to graph */
 
 	uint32_t port_user_data_size;	/**< extra size for port user data */
 
@@ -804,7 +819,7 @@ struct pw_impl_port_implementation {
 #define pw_impl_port_emit_param_changed(p,i)		pw_impl_port_emit(p, param_changed, 1, i)
 #define pw_impl_port_emit_latency_changed(p)		pw_impl_port_emit(p, latency_changed, 2)
 
-#define PW_IMPL_PORT_IS_CONTROL(port)	SPA_FLAG_MASK(port->flags, \
+#define PW_IMPL_PORT_IS_CONTROL(port)	SPA_FLAG_MASK((port)->flags, \
 						PW_IMPL_PORT_FLAG_BUFFERS|PW_IMPL_PORT_FLAG_CONTROL,\
 						PW_IMPL_PORT_FLAG_CONTROL)
 struct pw_impl_port {
